%% simulation
% This simulates the current beta_period with a range of beta values

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Create the current contact matrix
clear contact;
contact = beta_is_home(beta_period)*contact_home + beta_is_school(beta_period)*contact_school + ...
    beta_is_work(beta_period)*contact_work + beta_is_other(beta_period)*contact_other;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 % Manually adjust the age profile for transmission to 80+ (mu_e scaling)
 contact(3:5,5) = master_80con_scale*contact(3:5,5);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% Initial population conditions:
if (beta_period == 1)

    susceptible = zeros(num_classes,num_days);
    exposed = zeros(num_classes,num_days);
    infectious = zeros(num_classes,2,num_days);
    hospitalised = zeros(num_classes,2,num_days); % (:,:,1) = regular ward, (:,:,2) = ICU 
    recovered = zeros(num_classes,2,num_days);
    dead = zeros(num_classes,num_days);
    
    projected_hosp = zeros(num_classes,num_days); % new admissions to hospital
    projected_icu = zeros(num_classes,num_days); % new admissions to ICU
    projected_deaths = zeros(num_classes,num_days); % new deaths
    roll_projected_hosp = zeros(num_classes,num_days); % rolling average of new admissions to hospital
    residuals = zeros(num_classes,num_days);
    
    % Remaining members of population are susceptible
    for u1 = 1:1:num_classes
        susceptible(u1,1) = population*age_class_frac(u1)-infectious(u1,1)-infectious(u1,2)-exposed(u1)-recovered(u1,1)-recovered(u1,2);
    end

else
    % In any time period after the initial, we reload the final fitted
    % values from the best fit of the previous time period
    susceptible = best_susceptible;
    exposed = best_exposed;
    infectious = best_infectious;
    hospitalised = best_hospitalised;
    recovered = best_recovered;
    dead = best_dead;
    projected_hosp = best_projected_hosp;
    projected_icu = best_projected_icu;
    projected_deaths = best_projected_deaths;
    roll_projected_hosp = best_roll_projected_hosp;
    residuals = best_residuals;
end

% Reset residuals for this simulation
ss_res = 0.0;
pos_res = 0.0;

% Reset the time series over the period of this simulation
susceptible(:,beta_period_begin(beta_period)+1:end) = 0.0;
exposed(:,beta_period_begin(beta_period)+1:end) = 0.0;
infectious(:,:,beta_period_begin(beta_period)+1:end) = 0.0;
hospitalised(:,:,beta_period_begin(beta_period)+1:end) = 0.0;
recovered(:,:,beta_period_begin(beta_period)+1:end) = 0.0;
dead(:,beta_period_begin(beta_period)+1:end) = 0.0;

projected_hosp(:,beta_period_begin(beta_period)+1:end) = 0.0; % save the previous days for rolling-average
projected_icu(:,beta_period_begin(beta_period)+1:end) = 0.0;
projected_deaths(:,beta_period_begin(beta_period)+1:end) = 0.0;
roll_projected_hosp(:,beta_period_begin(beta_period)+1:end) = 0.0; % save the previous days for rolling-average
residuals(:,beta_period_begin(beta_period)+1:end) = 0.0;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Run the simulation

for i0 = beta_period_begin(beta_period)+1:1:beta_period_end(beta_period)+1+delay

    if (beta_period == 1)

        % The first infection begins
        if (i0 == initial_day+1)
            exposed(initial_class,i0-1) = 1.0;
            susceptible(initial_class,i0-1) = susceptible(initial_class,i0-1)-1.0;
        end

    end

    if (i0 > num_days) 
        % Check if projecting beyond the days for which we have empirical data
        
        break;
        
    end

    % Iterate the equations for this time-step
    run iterate
    
    % Collect that day's hospital admission prediction data
    for m1 = 1:1:num_classes
        projected_hosp(m1,i0) = temp_NEW_hospitalised(m1); % new hospital admissions
        projected_icu(m1,i0) = temp_NEW_icu(m1); % new icu admissions
        projected_deaths(m1,i0) = temp_dead(m1); % new deaths
    end

    % Calculate the 7-day rolling average of hospital admissions in each class
    if (i0 >= roll_ave_period) % start this at day 7

        for k1 = 0:1:roll_ave_period-1 % for 0-6 days prior
            for k2 = 1:1:num_classes
                roll_projected_hosp(k2,i0) = roll_projected_hosp(k2,i0)+projected_hosp(k2,i0-k1);
            end
        end
        % mean
        roll_projected_hosp(:,i0) = roll_projected_hosp(:,i0)/roll_ave_period; 
    end

    % Add this day's square residuals across age-classes to the ongoing cumulation
    if (collect_residuals == true)
        for l2 = 1:1:num_classes
            
            % store the actual residuals for resampling
            residuals(l2,i0) = roll_projected_hosp(l2,i0)-actual_daily_RA_hosp(l2,i0);
            ss_res = ss_res+(roll_projected_hosp(l2,i0)-actual_daily_RA_hosp(l2,i0)).^2.0;
            pos_res = pos_res+roll_projected_hosp(l2,i0)-actual_daily_RA_hosp(l2,i0);

            % if pos_res is large and positive, the current beta is too big.
            % if pos_res is large and negative, the current beta is too small.
        end
    end
    
    % Calculate current susceptible fraction of the population
    sus_frac = zeros(num_classes,1);
    for jj1 = 1:1:num_classes
        sus_frac(jj1) = susceptible(jj1,i0)/class_population(jj1);
    end
    
    % Check if we have exceeded the days for this time-period
    if (i0 == beta_period_end(beta_period)+1)
        
        % save the end point variables to resume at the beginning of the
        % next beta_period if this is the best fit for *this* beta_period
        saved_susceptible = susceptible;
        saved_exposed = exposed;
        saved_infectious = infectious;
        saved_hospitalised = hospitalised;
        saved_recovered = recovered;
        saved_dead = dead;
        saved_projected_hosp = projected_hosp;
        saved_projected_icu = projected_icu;
        saved_projected_deaths = projected_deaths;
        saved_roll_projected_hosp = roll_projected_hosp;
        saved_residuals = residuals;
        saved_ss_res = ss_res;
        saved_pos_res = pos_res;
        
    end  
end

% Check if the SS_res is better, and save if so
if (ss_res == min(ss_res,best_ss_res))

    % Note that the "best_" values, are the *current* best for this period,
    % and they are subject to change if a better fit is found with a
    % subsequent set of parameter choices:
    best_ss_res = ss_res;
    best_pos_res = pos_res;
    
    best_actual_ss_res = saved_ss_res;
    best_actual_pos_res = saved_pos_res;

    best_beta_store(beta_period) = beta;
    best_clinical_beta_scale = clinical_beta_scale;
    best_initial_day = initial_day;
    best_initial_class = initial_class;

    % Also store the final arrays of susceptible, exposed, etc. as these
    % will be needed for the next time period if these are the best.
    best_susceptible = saved_susceptible;
    best_exposed = saved_exposed;
    best_infectious = saved_infectious;
    best_hospitalised = saved_hospitalised;
    best_recovered = saved_recovered;
    best_dead = saved_dead;
    best_projected_hosp = saved_projected_hosp;
    best_projected_icu = saved_projected_icu;
    best_projected_deaths = saved_projected_deaths;
    best_roll_projected_hosp = saved_roll_projected_hosp;
    best_residuals = saved_residuals;
 
end

% Remove temporary variables
clear saved_susceptible saved_exposed saved_infectious saved_hospitalised saved_recovered saved_dead...
        saved_projected_hosp saved_projected_icu saved_projected_deaths saved_roll_projected_hosp;
