%% Main Script to determine best fit simulation
%
% This script executes a series of partial simulations, selecting the
% parameter values (day and age class of initial infection) and the 
% base transmission rate beta for each of the nine time intervals that 
% best fit the 7-day rolling average of the daily new hospital admissions
% in each of five 20-year age classes.


%% Setup
    
% Coupled SEIIRR model for Northern Ireland with 5 age-classes
clear;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% IDENTIFIER for this simulation:
version = 101;
version_text = 'Standard model variation used in main article.';

if exist([num2str(version)],'dir')
    % this folder and version number have already been used, confirm to 
    % overwrite the figures and workspace
    warning('This simulation number already exists.');
    while (true)
        resp = input('Continue anyway? [YES] ','s');
        if isempty(resp)
            continue;
        else
            if (strcmp(resp,'YES')==1)
                break;
            else
                clear;
                return;
            end
        end
    end
else
    % Create the directory if this version code has not been used before
    mkdir([num2str(version)]);
end
% Write a .txt file in the subdirectory, describing this version/scenario
version_file_path=sprintf('%d/version_description.txt',version);
fid = fopen(version_file_path,'wt');
fprintf(fid,'%s\n',version_text);
fclose(fid);

% These values are fitted across multiple simulations of the model:
master_80con_scale = 3.0; % mu_e
master_delay = 2; % D
master_con_work = 0.0; % mu_w
master_con_other = 0.95; % mu_o
subclinical_precedes_clinical = 1; % which model configuration do you want? Currently 1.
load_original_empirical_data = true; % set to false if you want to resample residuals
collect_residuals = true; % needs to be set to false if you want num_days>400

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Choose these:
beta_increment_num_1 = 100;
beta_increment_num_rem = 100;

clinical_beta_scale = 3*439/(4*561);
% i_C, chosen so that 56.1% of infections are due to subclinical class 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

plot_the_graphs = true;

% Setup: variables and files
num_classes = 5;
num_days = 400; % max is 381
roll_ave_period = 7;
num_beta_periods = 9;
delay = master_delay;
epsilon_min = 0.00001;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

euler_delta = 1.0; % the value of delta in the Euler method
sigma = 1/5.1; % incubation rate
gamma = zeros(2,1);
rate_leave_hosp = 1/11;
rate_leave_icu = 1/8;

% Time intervals:
beta_period_begin = [1,84,164,230,291,307,332,346,361]; % when does each time-period of different beta's begin?
beta_period_end = [83,163,229,290,306,331,345,360,num_days-1]; % when does it end?

% Scaling for each contact matrix during these periods:
beta_is_home   = [1,1,1,1,1,1,1,1,1];
beta_is_school = [1,0,0,1,0,1,1,1,0];
beta_is_work   = [1,0,master_con_work,master_con_work,master_con_work,master_con_work,0,master_con_work,0];
beta_is_other  = [1,master_con_other,master_con_other,master_con_other,master_con_other,master_con_other,master_con_other,master_con_other,master_con_other];

% Justification:
% 1: 1st January - 23rd March : Pre-COVID society (schools open)
% 84: 24th March - 11th June : Lockdown 1 (schools closed)
% 164: 12th June - 16th August : Shops re-open (schools closed)
% 230: 17th August - 16th October : Pupils begin return (schools open)
% 291: 17th October - 1st November : Month of restrictions (schools closed)
% 307: 2nd November - 26th November : Some restrictions continue but schools
% re-open (schools open)
% 332: 27th November - 10th December : Two-week circuit breaker but schools remain
% open (schools open)
% 346: 11th December - 25th December : Circuit breaker ends (schools open)
% 361: 26th December - end : 6-week lockdown beginning on Boxing Day. 
% VACCINE DISTRIBUTION BEGINS IN JANUARY 2021


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Read in data from files

% load the daily rolling average hospitalisations (by age class) to fit
temp1 = readmatrix('Data/hospital_rollave_PREPARED.csv');
actual_daily_RA_hosp = transpose(temp1);
clear temp1;


% Set total population sizes
population = readmatrix('Data/total_pop.csv');

% Set fraction of each regional population in each of the five age classes
age_class_frac = readmatrix('Data/age_fractions.csv');

% Set 10-year age class based susceptibility and symptomatic
age_sus_asym_hosp_icu_rates = readmatrix('Data/age_group_rates.csv');

% Set 5-year age class based hospitalisation, ICU and death rates
age_hosp_rates_5year = readmatrix('Data/age_group_hosp_rates_recent.csv');

% Set fractions of 5-year age classes (as a fraction of the five 20-year classes)
age_fractions_5year = readmatrix('Data/pop_fractions_5year.csv');

% Set fractions of 10-year age classes (as a fraction of the five 20-year classes)
age_fractions = readmatrix('Data/pop_fractions.csv');

% Set absolute numbers in each class
class_population = zeros(num_classes,1);
for j1 = 1:1:num_classes
    class_population(j1) = population*age_class_frac(j1);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Contact matrices:
% Load the NI-weighted spatially-invariant 5-age-class contact matrices for...

% HOME:
contact_home = readmatrix('Data/NI_contact_HOME.csv');

% SCHOOL:
contact_school = readmatrix('Data/NI_contact_SCHOOL.csv');

% WORK:
contact_work = readmatrix('Data/NI_contact_WORK.csv');

% OTHER:
contact_other = readmatrix('Data/NI_contact_OTHER.csv');

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Initialise the output storage:
output_beta_period= zeros(num_beta_periods,1);
output_initial_class = zeros(num_beta_periods,1);
output_initial_day = zeros(num_beta_periods,1);
output_clinical_beta_scale = zeros(num_beta_periods,1);
output_beta_store = zeros(num_beta_periods,1);
output_pos_res = zeros(num_beta_periods,1);
output_ss_res = zeros(num_beta_periods,1);
output_actual_ss_res = zeros(num_beta_periods,1);
output_actual_pos_res = zeros(num_beta_periods,1);
output_residuals = zeros(num_classes,num_days);

% Temporary holders of best time series while still being fitted:
best_beta_store = zeros(num_beta_periods,1);
best_susceptible = zeros(num_classes,num_days);
best_exposed = zeros(num_classes,num_days);
best_infectious = zeros(num_classes,2,num_days);
best_hospitalised = zeros(num_classes,2,num_days);
best_recovered = zeros(num_classes,2,num_days);
best_dead = zeros(num_classes,num_days);
best_residuals = zeros(num_classes,num_days);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Construct the age-weighted hospitalisation and death profile
age_hosp_profile = zeros(num_classes,1);
age_death_ward_profile = zeros(num_classes,1);
age_icu_profile = zeros(num_classes,1);
age_death_icu_profile = zeros(num_classes,1);

for n1 = 1:1:4 % 0-19 classes
   age_hosp_profile(1) = age_hosp_profile(1) + age_fractions_5year(n1)*0.75*age_hosp_rates_5year(n1,1);
   age_icu_profile(1) = age_icu_profile(1) + age_fractions_5year(n1)*0.24*age_hosp_rates_5year(n1,2);
   age_death_ward_profile(1) = age_death_ward_profile(1) + age_fractions_5year(n1)*0.46*age_hosp_rates_5year(n1,3);
   age_death_icu_profile(1) = age_death_icu_profile(1) + age_fractions_5year(n1)*(0.67*age_hosp_rates_5year(n1,4)...
       + (1.0-0.67*age_hosp_rates_5year(n1,4))*0.35*age_hosp_rates_5year(n1,5));
   % include deaths in "Stepdown" in general wards after leaving ICU
end
age_death_ward_profile(1) = age_death_ward_profile(1)*(1.0-age_icu_profile(1));
% the probability of death in general ward is applied to those who do
% NOT get triaged to ICU in the Imperial paper

for n2 = 5:1:8 % 20-39 classes
   age_hosp_profile(2) = age_hosp_profile(2) + age_fractions_5year(n2)*0.75*age_hosp_rates_5year(n2,1);
   age_icu_profile(2) = age_icu_profile(2) + age_fractions_5year(n2)*0.24*age_hosp_rates_5year(n2,2);
   age_death_ward_profile(2) = age_death_ward_profile(2) + age_fractions_5year(n2)*0.46*age_hosp_rates_5year(n2,3);   
   age_death_icu_profile(2) = age_death_icu_profile(2) + age_fractions_5year(n2)*(0.67*age_hosp_rates_5year(n2,4)...
       + (1.0-0.67*age_hosp_rates_5year(n2,4))*0.35*age_hosp_rates_5year(n2,5));
end
age_death_ward_profile(2) = age_death_ward_profile(2)*(1.0-age_icu_profile(2));

for n3 = 9:1:12 % 40-59 classes
   age_hosp_profile(3) = age_hosp_profile(3) + age_fractions_5year(n3)*0.75*age_hosp_rates_5year(n3,1);
   age_icu_profile(3) = age_icu_profile(3) + age_fractions_5year(n3)*0.24*age_hosp_rates_5year(n3,2);
   age_death_ward_profile(3) = age_death_ward_profile(3) + age_fractions_5year(n3)*0.46*age_hosp_rates_5year(n3,3);
   age_death_icu_profile(3) = age_death_icu_profile(3) + age_fractions_5year(n3)*(0.67*age_hosp_rates_5year(n3,4)...
       + (1.0-0.67*age_hosp_rates_5year(n3,4))*0.35*age_hosp_rates_5year(n3,5));
end
age_death_ward_profile(3) = age_death_ward_profile(3)*(1.0-age_icu_profile(3));

for n4 = 13:1:16 % 60-79 classes
   age_hosp_profile(4) = age_hosp_profile(4) + age_fractions_5year(n4)*0.75*age_hosp_rates_5year(n4,1);
   age_icu_profile(4) = age_icu_profile(4) + age_fractions_5year(n4)*0.24*age_hosp_rates_5year(n4,2);
   age_death_ward_profile(4) = age_death_ward_profile(4) + age_fractions_5year(n4)*0.46*age_hosp_rates_5year(n4,3);
   age_death_icu_profile(4) = age_death_icu_profile(4) + age_fractions_5year(n4)*(0.67*age_hosp_rates_5year(n4,4)...
       + (1.0-0.67*age_hosp_rates_5year(n4,4))*0.35*age_hosp_rates_5year(n4,5));
end
age_death_ward_profile(4) = age_death_ward_profile(4)*(1.0-age_icu_profile(4));

for n5 = 17:1:17 % 80+ classes
   age_hosp_profile(5) = 0.75*age_hosp_rates_5year(n5,1);
   age_icu_profile(5) = 0.24*age_hosp_rates_5year(n5,2);
   age_death_ward_profile(5) = 0.46*age_hosp_rates_5year(n5,3)*(1.0-age_icu_profile(5));
   age_death_icu_profile(5) = 0.67*age_hosp_rates_5year(n5,4)+(1.0-0.67*age_hosp_rates_5year(n5,4))*0.35*age_hosp_rates_5year(n5,5);
end


% Set-up susceptibility_delta and symptomatic epsilon based on the values for the smaller 10-year bands:
epsilon = zeros(num_classes,1);
susceptibility_delta = zeros(num_classes,1);

for n1 = 1:1:2 % 0-19 classes
    susceptibility_delta(1) = susceptibility_delta(1) + age_fractions(n1)*age_sus_asym_hosp_icu_rates(n1,1);
    epsilon(1) = epsilon(1) + age_fractions(n1)*age_sus_asym_hosp_icu_rates(n1,2);
end
for n2 = 3:1:4 % 20-39 classes
    susceptibility_delta(2) = susceptibility_delta(2) + age_fractions(n2)*age_sus_asym_hosp_icu_rates(n2,1);
    epsilon(2) = epsilon(2) + age_fractions(n2)*age_sus_asym_hosp_icu_rates(n2,2);
end
for n3 = 5:1:6 % 40-59 classes
    susceptibility_delta(3) = susceptibility_delta(3) + age_fractions(n3)*age_sus_asym_hosp_icu_rates(n3,1);
    epsilon(3) = epsilon(3) + age_fractions(n3)*age_sus_asym_hosp_icu_rates(n3,2);
end
for n4 = 7:1:8 % 60-79 classes
    susceptibility_delta(4) = susceptibility_delta(4) + age_fractions(n4)*age_sus_asym_hosp_icu_rates(n4,1);
    epsilon(4) = epsilon(4) + age_fractions(n4)*age_sus_asym_hosp_icu_rates(n4,2);
end
for n5 = 9:1:9 % 80+ classes
    susceptibility_delta(5) = susceptibility_delta(5) + age_sus_asym_hosp_icu_rates(n5,1);
    epsilon(5) = epsilon(5) + age_sus_asym_hosp_icu_rates(n5,2);
end


% Choose which model configuration you want:
gamma(1) = 1/3;
if (subclinical_precedes_clinical == 1)
    % This is the model version where every case becomes subclinical and
    % infectious for 3 days, but then only a fraction become clinical
    % infectious for a further 4 days.
   
    gamma(2) = 1/4;
    exposed_to_sub = ones(num_classes,1);
    exposed_to_clinical = zeros(num_classes,1);
    sub_to_clinical = epsilon;
    sub_to_recovery = ones(num_classes,1) - epsilon;

else
    % This is the version where after 5.1 days of being exposed, cases
    % either become subclinical infectious for 3 days OR clinical
    % infectious for 7 days.
    
    gamma(2) = 1/7;
    exposed_to_sub = ones(num_classes,1) - epsilon;
    exposed_to_clinical = epsilon;
    sub_to_clinical = zeros(num_classes,1);
    sub_to_recovery = ones(num_classes,1);
    
    % Note that if clinical_beta_scale = 1 then these two versions
    % are not going to be distinct in the overall progression to recovery
    % and hospitalisation etc.
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   
%% First beta period


% Fit the transmission rate beta in the initial pre-pandemic period before
% the first lockdown (1st January - 22nd March 2020)

beta_period = 1;
best_ss_res = 1000000000.0;
lower_beta = 0.0;

% IDENTIFY A SUITABLE UPPER BOUND for Beta_0
upper_beta = 0.0;

% set initial conditions to MINIMISE infection - set a single exposed case on January 1st
% this will see how big beta would have to be in the worst case to get some hospitalisations
initial_class = 4; % compromise between higher contact and higher IFR
initial_day = 31;

% Identify the correct order of beta_1
for m3 = -15:1:30
    beta = 10.^m3;
    
    fprintf('Checking beta_1 order: %d\n',m3);
    run Code/simulation

    if (pos_res > 10000.0)
        % first instance of getting too big
        upper_beta = 10.^m3; 
        save_m3 = m3-1;
        break
    end

end

% Now tune the best beta_1 and fit the initial conditions
best_ss_res = 1000000000.0;
increment = (upper_beta-lower_beta)/beta_increment_num_1;

% loop which class the first exposed case is in
for initial_class = 1:1:num_classes

	% loop which day they appear
    for initial_day = 0:1:31

        fprintf('Testing initial_class = %d;  initial_day = %d\n',initial_class,initial_day);
        for beta = lower_beta:increment:upper_beta

			% Run the simulation for this first interval with these
			% parameters
		    run Code/simulation
		    
        end
    end
end

% For this first beta_period, save the overall best SS_res that was achieved and
% the corresponding parameter choices
output_beta_period(1) = beta_period;
output_initial_class(1) = best_initial_class;
output_initial_day(1) = best_initial_day;
output_clinical_beta_scale(1) = best_clinical_beta_scale;
output_beta_store(1) = best_beta_store(beta_period);
output_pos_res(1) = best_pos_res;
output_ss_res(1) = best_ss_res;
output_actual_ss_res(1) = best_actual_ss_res;
output_actual_pos_res(1) = best_actual_pos_res;

% Set subsequent parameters to the best ones:
initial_day = best_initial_day;
initial_class = best_initial_class;


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Remaining beta periods

% Iterate over the 8 remaining time-periods, performing multiple partial
% simulations for each to fit the corresponding transmission rate beta
% during this interval
for beta_period = 2:1:num_beta_periods
    
    fprintf('Fitting beta period = %d\n',beta_period);

	% reset the fitting values
    best_ss_res = 1000000000.0; % start with a very large value
   	
  	% the range of beta values searched should depend on the time-period

  	if (beta_period == 2) % lockdown 1

  		lower_beta = 0.0;
  		upper_beta = 1.5*best_beta_store(1);

    else
        
  		lower_beta = 0.1*best_beta_store(2); % less strict than Lockdown 1
  		upper_beta = 1.5*best_beta_store(1); % less relaxed than pre-lockdown
        % Note that these bounds aren't too meaningful, as a lot of the
        % spread will instead be controlled by the contact matrices
        % switching on and off.
        % Variances in this would account for things that control the
        % inherent infectiousness of a contact, such as:
        % i) sanitiser use
        % ii) social distancing
        % iii) masks
        % iv) more infectious variants of the virus
        
    end

  
    % Increment the beta values
    increment = (upper_beta-lower_beta)/beta_increment_num_rem;
    for beta = lower_beta:increment:upper_beta

        % Run the simulation
        run Code/simulation

    end

	% For this beta_period, save the final best SS_res and the corresponding
    % best parameter choices to the output_ variables. These are the "true
    % best" values, no longer subject to change:
    output_beta_period(beta_period) = beta_period;
    output_initial_class(beta_period) = best_initial_class;
    output_initial_day(beta_period) = best_initial_day;
    output_clinical_beta_scale(beta_period) = best_clinical_beta_scale;
    output_beta_store(beta_period) = best_beta_store(beta_period);
    output_pos_res(beta_period) = best_pos_res;
    output_ss_res(beta_period) = best_ss_res;
    % this is the sum of the square residuals created while actually doing the fit, but it means that the 
    % "delay" days are counted twice - (a) as part of the delayed previous
    % period, and (b) as part of the subsequent period.
    % This results in a bias against larger delays, and thus can't be
    % used for comparison between different delays.
  
    % The following is the sum of square residuals between the empirical
    % and the final projected data. I would hope that minimising ss_res
    % will lead to minimising actual_ss_res more effectively - that is the
    % point!
 
    output_actual_ss_res(beta_period) = best_actual_ss_res;
    output_actual_pos_res(beta_period) = best_actual_pos_res;
    
end	

% Print the sum of the square residuals for the final, best-fit simulation
sum_of_square_residuals = sum(output_actual_ss_res); % 

% Calculate the R-squared coefficient of determination:
mean_RA = mean(actual_daily_RA_hosp(:,1:400),'all');
ss_tot = 0.0;
for i=1:1:num_classes
    for j=1:1:num_days
        ss_tot = ss_tot + (actual_daily_RA_hosp(i,j) - mean_RA)^2;
    end
end
r_squared = 1 - sum_of_square_residuals/ss_tot;
fprintf('R^2 coefficient of determination = %d\n',r_squared);

% For full time-series of the residuals, output_ is just the same
% as best_
output_residuals = best_residuals; 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Now put the full time-series together for the complete fitted simulation

% Age-structured by amalgamated wider groups:
% all_infected is all individuals who are subclinical infectious, clinical
% infectious, hospitalised, or in ICU
all_infected = zeros(num_classes,num_days);
% all_hospitalised is all individuals who are in hospital or ICU
all_hospitalised = zeros(num_classes,num_days);
% all_recovered is all individuals recovered from either a subclinical
% infection or a clinical infection
all_recovered = zeros(num_classes,num_days);
for m0=1:1:num_classes
   all_infected(m0,:) =  best_infectious(m0,1,:) + best_infectious(m0,2,:) + best_hospitalised(m0,1,:) + best_hospitalised(m0,2,:);
   all_hospitalised(m0,:) = best_hospitalised(m0,1,:) + best_hospitalised(m0,2,:);
   all_recovered(m0,:) = best_recovered(m0,1,:) + best_recovered(m0,2,:);
end    


% Now combine data to obtain time-series of the compartments for the entire
% population. For example, susceptible_full_total is the time-series of the
% total number of Susceptible individuals of ANY age:

% Totals of each compartment
susceptible_full_total = zeros(num_days,1);
exposed_full_total = zeros(num_days,1);
infectious_full_total = zeros(2,num_days);
hospitalised_full_total = zeros(2,num_days);
recovered_full_total = zeros(2,num_days);
dead_full_total = zeros(num_days,1);

% Totals of projections of new daily events
projected_hosp_full_total = zeros(num_days,1);
projected_icu_full_total = zeros(num_days,1);
projected_deaths_full_total = zeros(num_days,1);

% Totals of wider groups
all_infected_full_total = zeros(num_days,1);
all_recovered_full_total = zeros(num_days,1);
all_hospitalised_full_total = zeros(num_days,1);

% Total of empirical data
actual_daily_RA_hosp_full_total = zeros(num_days,1);
for t1 = 1:1:num_classes
    actual_daily_RA_hosp_full_total = actual_daily_RA_hosp_full_total+transpose(squeeze(actual_daily_RA_hosp(t1,1:num_days)));
end


for i0=1:1:num_days
    for i1 = 1:1:num_classes

        % Compartments
        susceptible_full_total(i0) = susceptible_full_total(i0) + best_susceptible(i1,i0);
        exposed_full_total(i0) = exposed_full_total(i0) + best_exposed(i1,i0);
        for i3=1:1:2
            infectious_full_total(i3,i0) = infectious_full_total(i3,i0) + best_infectious(i1,i3,i0);
            hospitalised_full_total(i3,i0) = hospitalised_full_total(i3,i0) + best_hospitalised(i1,i3,i0);
        end
        dead_full_total(i0) = dead_full_total(i0) + best_dead(i1,i0);

        % Daily new projections
        projected_hosp_full_total(i0) = projected_hosp_full_total(i0) + best_projected_hosp(i1,i0);
        projected_icu_full_total(i0) = projected_icu_full_total(i0) + best_projected_icu(i1,i0);
        projected_deaths_full_total(i0) = projected_deaths_full_total(i0) + best_projected_deaths(i1,i0);

        % Amalgamated compartments
        all_infected_full_total(i0) = all_infected_full_total(i0) + all_infected(i1,i0);
        all_hospitalised_full_total(i0) = all_hospitalised_full_total(i0) + all_hospitalised(i1,i0);
        all_recovered_full_total(i0) = all_recovered_full_total(i0) + all_recovered(i1,i0);

    end
end

if (plot_the_graphs == true)
    run Code/graph_plotting
end

% Save workspace to the version-specific directory
version_workspace_path = sprintf('%d/version_workspace_%d.mat',version,version);
save(version_workspace_path);
